# Generated by cpp11: do not edit by hand

df_parse_sas_file <- function(spec_b7dat, spec_b7cat, encoding, catalog_encoding, cols_skip, n_max, rows_skip, name_repair) {
  .Call(`_haven_df_parse_sas_file`, spec_b7dat, spec_b7cat, encoding, catalog_encoding, cols_skip, n_max, rows_skip, name_repair)
}

df_parse_sas_raw <- function(spec_b7dat, spec_b7cat, encoding, catalog_encoding, cols_skip, n_max, rows_skip, name_repair) {
  .Call(`_haven_df_parse_sas_raw`, spec_b7dat, spec_b7cat, encoding, catalog_encoding, cols_skip, n_max, rows_skip, name_repair)
}

df_parse_xpt_file <- function(spec, cols_skip, n_max, rows_skip, name_repair) {
  .Call(`_haven_df_parse_xpt_file`, spec, cols_skip, n_max, rows_skip, name_repair)
}

df_parse_xpt_raw <- function(spec, cols_skip, n_max, rows_skip, name_repair) {
  .Call(`_haven_df_parse_xpt_raw`, spec, cols_skip, n_max, rows_skip, name_repair)
}

df_parse_dta_file <- function(spec, encoding, cols_skip, n_max, rows_skip, name_repair) {
  .Call(`_haven_df_parse_dta_file`, spec, encoding, cols_skip, n_max, rows_skip, name_repair)
}

df_parse_dta_raw <- function(spec, encoding, cols_skip, n_max, rows_skip, name_repair) {
  .Call(`_haven_df_parse_dta_raw`, spec, encoding, cols_skip, n_max, rows_skip, name_repair)
}

df_parse_sav_file <- function(spec, encoding, user_na, cols_skip, n_max, rows_skip, name_repair) {
  .Call(`_haven_df_parse_sav_file`, spec, encoding, user_na, cols_skip, n_max, rows_skip, name_repair)
}

df_parse_sav_raw <- function(spec, encoding, user_na, cols_skip, n_max, rows_skip, name_repair) {
  .Call(`_haven_df_parse_sav_raw`, spec, encoding, user_na, cols_skip, n_max, rows_skip, name_repair)
}

df_parse_por_file <- function(spec, encoding, user_na, cols_skip, n_max, rows_skip, name_repair) {
  .Call(`_haven_df_parse_por_file`, spec, encoding, user_na, cols_skip, n_max, rows_skip, name_repair)
}

df_parse_por_raw <- function(spec, encoding, user_na, cols_skip, n_max, rows_skip, name_repair) {
  .Call(`_haven_df_parse_por_raw`, spec, encoding, user_na, cols_skip, n_max, rows_skip, name_repair)
}

write_sav_ <- function(data, path, compress) {
  invisible(.Call(`_haven_write_sav_`, data, path, compress))
}

write_dta_ <- function(data, path, version, label, strl_threshold) {
  invisible(.Call(`_haven_write_dta_`, data, path, version, label, strl_threshold))
}

write_sas_ <- function(data, path) {
  invisible(.Call(`_haven_write_sas_`, data, path))
}

write_xpt_ <- function(data, path, version, name, label) {
  invisible(.Call(`_haven_write_xpt_`, data, path, version, name, label))
}
